import java.util.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

public class IRKeypad 
  extends JPanel 
  implements ActionListener {
  
  private Map commands_; // command name -> IR code

  public IRKeypad() {
    super();
    listeners_ = new Vector();
    commands_ = new HashMap();

    addCommands(commands_);

    setLayout(new GridLayout(6, 5));
    
    Set keys = commands_.keySet();
    for (Iterator i = keys.iterator(); i.hasNext(); ) {
      String cmd = (String) i.next();
      JButton b = new JButton(cmd);
      b.addActionListener(this);
      add(b);
    }
  }

  public static interface IRListener {
    public void keyPressed(int code);
  }

  private Vector listeners_;
  public void addListener(IRListener l) {
    listeners_.add(l);
  }
    
  private void notifyListeners(int code) {
    for (int i = 0; i < listeners_.size(); i++)
      ((IRListener) listeners_.get(i)).keyPressed(code);
  }

  public void actionPerformed(ActionEvent e) { 
    String cmdName = e.getActionCommand();
    Integer i = (Integer) commands_.get(cmdName);
    Util.assert(i != null);
    notifyListeners(i.intValue());      
  }

  private void addCommands(Map m) {
    m.put("1"   , new Integer(0xf786));
    m.put("2"	, new Integer(0xf746));
    m.put("3"	, new Integer(0xf7c6));
    m.put("4"	, new Integer(0xf726));
    m.put("5"	, new Integer(0xf7a6));
    m.put("6"	, new Integer(0xf766));
    m.put("7"	, new Integer(0xf7e6));
    m.put("8"	, new Integer(0xf716));
    m.put("9"	, new Integer(0xf796));
    m.put("0"	, new Integer(0xf776));
    m.put("pip"	, new Integer(0xf7f6));
    m.put("back"	, new Integer(0xf70e));
    m.put("play"	, new Integer(0xf732));
    m.put("fwd"	, new Integer(0xf76e));
    m.put("rec"	, new Integer(0xf743));
    m.put("stop"	, new Integer(0xf7c2));
    m.put("pause"	, new Integer(0xf7b2));
    m.put("disp"	, new Integer(0xf703));
    m.put("sleep"	, new Integer(0xf7b3));
    m.put("guide"	, new Integer(0xf7b6));
    m.put("up"	, new Integer(0xf70b));
    m.put("left"	, new Integer(0xf74b));
    m.put("right"	, new Integer(0xf7cb));
    m.put("down"	, new Integer(0xf78b));
    m.put("menu"	, new Integer(0xf783));
    m.put("ok"	, new Integer(0xf72b));
    m.put("chup"	, new Integer(0xf70d));
    m.put("chdown"	, new Integer(0xf78d));
    m.put("recall"	, new Integer(0xf7ab));
    m.put("power"	, new Integer(0xf702));
  }
}
